// arch/x86/misc.cc -- Misc. arch-specific stuff
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/types.h>
#include <kern/libc.h>
#include <kern/arch.h>
#include <kern/i8259.h>
#include <kern/time.h>
#include <kern/thread.h>
#include <kern/mem.h>
#include <kern/process.h>
#include <kern/orb.h>

#include <arch/addrs.h>
#include <arch/multiboot.h>
#include <arch/regs.h>

extern u32 x86_init_ptbl_l2[1024];

struct X86ExTable {
	ulong faultaddr, handler;
};

extern X86ExTable extable_start, extable_end;

struct X86InitStack {
	u8 stack[4096 - ::Threads::thread_size];
	::Threads::Thread thread;
} __attribute__((aligned(4096))) x86_init_stack;

namespace Arch {
	namespace Priv {
		void set_idt();
		
		void show_regs(Regs *regs)
		{
			u32 esp = regs->cs & 3 ? regs->user_esp :
			          (u32)regs + (u32)sizeof(regs);

			printf("eax: 0x%08x   ecx: 0x%08x   edx: 0x%08x   ebx: 0x%08x\n"
			       "esp: 0x%08x   ebp: 0x%08x   esi: 0x%08x   edi: 0x%08x\n"
			       "eflags: 0x%08x ds: %04x es: %04x ss: %04x\n",
			       regs->eax, regs->ecx, regs->edx, regs->ebx, esp,
			       regs->ebp, regs->esi, regs->edi, regs->eflags,
			       regs->ds, regs->es, regs->cs & 3 ? regs->user_ss : regs->ds);

			printf("Stack trace:          ");
			u32 *frame = (u32 *)regs->ebp;
			
			for (int i = 2; i < 32; i++) {
				if ((u32)frame < 0x80000000UL)
					break;

				u32 stackptr = frame[1];
				frame = (u32 *)frame[0];

				if (!(i % 7))
					printf("\n");
				
				printf("0x%08x ", stackptr);
			}

			printf("\nStack dump:           ");

			for (int i = 2; i < 32; i++) {
				if (!(i % 7))
					printf("\n");
				
				printf("0x%08x ", ((u32 *)esp)[i-2]);
			}
			
		}

		struct TimerInt : public IRQ::Interrupt {
			bool action()
			{
				Time::monotonic_timers->run();
				return true;
			}
		};
		
		TimerInt timer_int;
		
		bool check_extable(Regs *regs, int cause)
		{
			X86ExTable *ext;
			assert((regs->cs & 3) == 0);
			
			for (ext = &extable_start; ext < &extable_end; ext++) {
				if (ext->faultaddr == regs->eip) {
					regs->eip = ext->handler;
					regs->edx = cause;
					return true;
				}
			}
		
			return false;
		}
	}

	using IRQ::i8259;
	::Threads::Thread *init_thread;
	
	void arch_init()
	{
		init_thread = &x86_init_stack.thread;
		Priv::early_adjust_mappings();
		Priv::set_idt();
		Priv::MultiBoot::process_info();
		i8259.init();
		
		u32 tss_addr = reinterpret_cast<u32>(&Priv::tss);
		x86_gdt[3].base_low = tss_addr & 0xffff;
		x86_gdt[3].base_mid = (tss_addr & 0xff0000) >> 16;
		x86_gdt[3].base_high = (tss_addr & 0xff000000) >> 24;
		
		Priv::tss.ss0 = 8;
		asm volatile("ltr %w0" : : "r" (0x18) : "memory");
		init_thread->aspace = new Mem::ProcAddrSpace(x86_init_ptbl_l2);
		init_thread->active_aspace = init_thread->aspace;
	}

	void timer_init()
	{
		IRQ::InterruptSlot *timer = i8259.get_slot(0);
		i8259.request_int(timer, &Priv::timer_int);
	}
}

using Arch::Priv::show_regs;
using Arch::Priv::Regs;

extern "C" void x86_do_diverr(Regs *regs)
{
	printf("Division error at 0x%04x:0x%08x\n", regs->cs, regs->eip);
	show_regs(regs);
	for(;;);
}

extern "C" void x86_do_debug(Regs *regs)
{
	printf("Debug exception at 0x%04x:0x%08x\n", regs->cs, regs->eip);
	show_regs(regs);
	for(;;);
}

extern "C" void x86_do_breakpoint(Regs *regs)
{
	printf("Breakpoint at 0x%04x:0x%08x\n", regs->cs, regs->eip);
	show_regs(regs);
	for(;;);
}

extern "C" void x86_do_invalid_insn(Regs *regs)
{
	printf("Invalid instruction at 0x%04x:0x%08x\n", regs->cs, regs->eip);
	show_regs(regs);
	for(;;);
}

extern "C" void x86_do_page_fault(Regs *regs, u32 fault_addr, u32 error_code)
{
	Mem::AddrSpace *as;
	int cause;

	if (in_fault)
		for(;;);

	// A reserved bit was set in the PTE; this is always a bug.
	if (error_code & 8)
		goto bad_fault;
	
	// Don't try to fix up a page fault if interrupts were disabled.  It is an
	// error to access non-locked pages with interrupts disabled.  Trying to
	// fix it up in the case of an access that would be legitimate if interrupts
	// were enabled would simply mask the loss of atomicity, and trying to grab
	// locks to look up the address if it is a completely bad reference won't
	// accomplish much other than decreasing the odds that the fault message
	// gets out.
	
	if (!(regs->eflags & 0x200))
		goto bad_fault;
	
	// Don't allow fault-ins using a borrowed addr-space.
	as = curthread->aspace;

	if (!as || curthread == Arch::init_thread)
		goto bad_fault;
	
	ll_ints_on();
	
	// FIXME: no-exec
	cause = as->handle_fault(fault_addr, error_code & 2,
	                         false /* error_code & 16 */,
	                         error_code & 4);
	
	if (cause < 0)
		return;
	
	if (!(error_code & 4) && check_extable(regs, cause))
		return;
	
	// FIXME: throw exception to user

bad_fault:
	ll_ints_off();
	in_fault++;
	
	printf("Page fault at 0x%04x:0x%08x for 0x%08x, error code: 0x%04x\n",
	       regs->cs, regs->eip, fault_addr, error_code);
	
	show_regs(regs);

	for(;;);
}

extern "C" void x86_do_gpf(Regs *regs, u32 error_code)
{
	if (in_fault)
		for(;;);
	
	in_fault++;

	printf("General protection fault at 0x%04x:0x%08x, error code: 0x%04x\n",
	       regs->cs, regs->eip, error_code);

	show_regs(regs);
	
	for(;;);
}

extern "C" void x86_do_irq(int irq)
{
	IRQ::i8259.handle_irq(irq - 0x20);
}

namespace Arch {
namespace Priv {
	struct OrbRegs {
		u32 ds, es;
	
		union {
			System::RunTime::ParamInfoBlock *pib;
			uintptr_t exptr;
		};
		
		union {
			uintptr_t caller;
			size_t exlen;
		};
		
		ulong eip, cs, eflags, user_esp, user_ss;
	};
}}

extern "C" void x86_invoke_method(Arch::Priv::OrbRegs *regs)
{
	assert(regs->cs & 3);
	printf("x86_invoke_method %p\n", regs->pib);
	
	try {
		ORB::invoke_method(regs->pib, regs->user_esp);
	}
	
	catch (SystemException &se) {
		// Copy exception to user
		// If an exception is thrown during dispatch, it must be
		// with the caller's address space, and user_esp must be
		// the caller's.
		
		printf("sys exception\n");
		
		typedef System::Exceptions::NativeCodeExceptionOriginInfo OriginInfo;
		OriginInfo *oi = OriginInfo::downcast(se.origin);
	
		if (oi)
			printf("PC %llx\n", oi->pc);
	}
	
	catch (...) {
		printf("other exception\n");
	}
}

extern "C" void x86_return_from_method(Arch::Priv::OrbRegs *regs)
{
	assert(regs->cs & 3);
	regs->user_esp = ORB::return_from_method(regs->exptr, regs->exlen);
}
