// mem/rmap.cc -- Reverse mapping from physical page frames (or 
// intermediate address spaces) to mappers.
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors or contributors be held liable for any damages
// arising from the use of this software.
// 
// Permission is hereby granted to everyone, free of charge, to use, copy,
// modify, prepare derivative works of, publish, distribute, perform,
// sublicense, and/or sell copies of the Software, provided that the above
// copyright notice and disclaimer of warranty be included in all copies or
// substantial portions of this software.

#include <kern/mem.h>
#include <kern/pagealloc.h>
#include <kern/rmap.h>
#include <kern/pagetable.h>
#include <kern/sched.h>
#include <kern/thread.h>

namespace Mem {
	static const ulong rmap_node_len =
		1 << ll_get_order_round_up(sizeof(RMapNode));
	
	void RMapTable::map(VirtualArea *dsva, PageTable *usptbl,
		                 u64 dsvaddr, u64 usvaddr)
	{
		RMapNode *dsrmap = dsva->aspace->page_table->
		                   rmap_table.lookup(dsvaddr, true);

		assert(!dsrmap->va);
		dsrmap->va = dsva;
		dsrmap->vaddr = page_align(dsvaddr);
		dsrmap->head.init();
		dsrmap->tail.init();
		
		if (usptbl) {
			RMapNode *usrmap = usptbl->rmap_table.lookup(usvaddr);
			assert(usrmap);
			assert(usrmap->va->aspace->page_table == usptbl);

			usrmap->head.add_front(&dsrmap->head);
		} else {
			// FIXME: If it ends up being useful, link into the phys-page
			// rmap list.
		}

		dsrmap->head.add_front(&dsrmap->tail);
	}

	void RMapTable::unmap(u64 virtaddr)
	{
		Lock::AutoLock autolock(rmap_lock);
		RMapNode *head = lookup(virtaddr);
		
		if (!head || !head->va)
			return;

		assert(head->vaddr == virtaddr);

		Util::ListNoAutoInit *node = &head->head, *oldnode;
		
		do {
			ulong off = reinterpret_cast<ulong>(node) & (rmap_node_len - 1);
			if (off == RMap::head_offset) {
				RMapNode *rmap = node->listentry(RMapNode, head);
			
				Region region = { rmap->vaddr,
				                  rmap->vaddr + Arch::page_shift - 1 };

				rmap->va->aspace->page_table->unmap(region);
				rmap->va = NULL;
			} else {
				assert(off == RMap::tail_offset);
			}

			oldnode = node;
			node = node->next;
			oldnode->del();
		} while (node != &head->tail);
		
		node->del();
	}

	void RMapTable::break_copy_on_write(u64 virtaddr, Page *new_page)
	{
		assert(rmap_lock.held_by_curthread());
		RMapNode *head = lookup(virtaddr);
		RMapNode *still_cow = NULL;
		
		assert(head && head->va);
		assert(head->vaddr == virtaddr);
		
		// If there are sibling or upstream mappings of this page,
		// detach the rmap list.
		
		if (head->head.prev != &head->tail) {
			head->head.prev->next = head->tail.next;
			head->tail.next->prev = head->head.prev;
			
			head->head.prev = &head->tail;
			head->tail.next = &head->head;
		}

		assert(head->tail.next == &head->head);
		Util::ListNoAutoInit *node = &head->head;
		
		do {
			ulong off = reinterpret_cast<ulong>(node) & (rmap_node_len - 1);
			if (off == RMap::head_offset) {
				RMapNode *rmap = node->listentry(RMapNode, head);
				RegionWithOffset region;
				
				region.start = rmap->vaddr;
				region.end = rmap->vaddr + Arch::page_shift - 1;
				region.offset = page_to_phys(new_page);
				
				PTEFlags flags = rmap->va->flags;
				
				// The faulting mapping always has PTE FaultOnWrite cleared;
				// downstream mappings have PTE FaultOnWrite cleared if they
				// are not downstream of different mapping with VA
				// FaultOnWrite set.  Downstream mappings should never have
				// PTE FaultOnWrite clear if VA FaultOnWrite is set; if the
				// downstream mapping had been cow-broken, it would have been
				// removed from this physpage's rmap list.
				
				if (flags.FaultOnWrite && node != &head->head && !still_cow)
					still_cow = rmap;
				
				if (still_cow)
					flags.FaultOnWrite = 1;
				else
					flags.FaultOnWrite = 0;
				
				rmap->va->aspace->page_table->map(region, flags);
			} else {
				assert(off == RMap::tail_offset);
				
				if (still_cow) {
					RMapNode *rmap = node->listentry(RMapNode, tail);

					// We've finished the downstreams of a COW mapping,
					// so stop marking pages as COW.

					if (rmap == still_cow)
						still_cow = NULL;
				}
			}

			node = node->next;
		} while (node != &head->tail);
		
		assert(!still_cow);
	}

	Lock::Lock rmap_lock;
}
